(** This file includes basic examples that each demonstrate a unique feature of
multiparty dependent separation protocols and Multris. *)
From multi_actris.channel Require Import proofmode.
Set Default Proof Using "Type".

Definition iProto_binders `{!heapGS Σ} : list (iProto Σ) :=
  [(<(Send,1)> MSG #20%Z ; END)%proto;
   (<(Recv,0) @ (x:Z)> MSG #x {{ ⌜(x > 0)%Z⌝ }} ; END)%proto].

Lemma iProto_consistent_binders `{!heapGS Σ} :
  ⊢ iProto_consistent iProto_binders.
Proof. iProto_consistent_take_steps. Qed.

Definition roundtrip_prog : val :=
  λ: <>,
     let: "cs" := new_chan #3 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     Fork (let: "x" := recv "c1" #0 in send "c1" #2 ("x"+#20));;
     Fork (let: "x" := recv "c2" #1 in send "c2" #0 ("x"+#30));;
     send "c0" #1 #100;; recv "c0" #2.

Section channel.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_roundtrip : list (iProto Σ) :=
     [(<(Send, 1) @ (x:Z)> MSG #x ; <(Recv, 2)> MSG #(x+50); END)%proto;
      (<(Recv, 0) @ (x:Z)> MSG #x ; <(Send, 2)> MSG #(x+20); END)%proto;
      (<(Recv, 1) @ (x:Z)> MSG #x ; <(Send, 0)> MSG #(x+30); END)%proto].

  Lemma iProto_roundtrip_consistent :
    ⊢ iProto_consistent iProto_roundtrip.
  Proof.
    rewrite /iProto_roundtrip. iProto_consistent_take_steps.
    replace (x + 20 + 30)%Z with (x+50)%Z by lia.
    iSplit; [done|]. iSplit; [done|]. 
    iProto_consistent_take_steps.
  Qed.

  Lemma roundtrip_prog_spec :
    {{{ True }}} roundtrip_prog #() {{{ RET #150 ; True }}}.
  Proof using chanG0 heapGS0 Σ.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_roundtrip with iProto_roundtrip_consistent
      as (c0 c1 c2) "Hc0" "Hc1" "Hc2".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>". wp_recv (x) as "_". wp_send with "[//]". done. }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>". wp_recv (x) as "_". wp_send with "[//]". done. }
    wp_send with "[//]". wp_recv as "_". by iApply "HΦ".
  Qed.

End channel.

Definition roundtrip_ref_prog : val :=
  λ: <>,
     let: "cs" := new_chan #3 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     Fork (let: "l" := recv "c1" #0 in "l" <- !"l" + #20;; send "c1" #2 "l");;
     Fork (let: "l" := recv "c2" #1 in "l" <- !"l" + #30;; send "c2" #0 #());;
     let: "l" := ref #100 in send "c0" #1 "l";; recv "c0" #2;; !"l".

Section roundtrip_ref.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_roundtrip_ref : list (iProto Σ) :=
    [(<(Send, 1) @ (l:loc) (x:Z)> MSG #l {{ (l ↦ #x)%I }} ;
            <(Recv, 2)> MSG #() {{ l ↦ #(x+50) }} ; END)%proto;
     (<(Recv, 0) @ (l:loc) (x:Z)> MSG #l {{ (l ↦ #x)%I }} ;
            <(Send, 2)> MSG #l {{ l ↦ #(x+20) }}; END)%proto;
     (<(Recv, 1) @ (l:loc) (x:Z)> MSG #l {{ (l ↦ #x)%I }} ;
            <(Send, 0)> MSG #() {{ l ↦ #(x+30) }}; END)%proto].

  Lemma iProto_roundtrip_ref_consistent :
    ⊢ iProto_consistent iProto_roundtrip_ref.
  Proof.
    rewrite /iProto_roundtrip_ref.
    iProto_consistent_take_steps.
    replace (x0 + 20 + 30)%Z with (x0+50)%Z by lia. iFrame.
    iProto_consistent_take_step.
  Qed.

  Lemma roundtrip_ref_prog_spec :
    {{{ True }}} roundtrip_ref_prog #() {{{ RET #150 ; True }}}.
  Proof using chanG0.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_roundtrip_ref with iProto_roundtrip_ref_consistent
      as (c0 c1 c2) "Hc0" "Hc1" "Hc2".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>".
      wp_recv (l x) as "Hl". wp_load. wp_store. by wp_send with "[$Hl]". }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>".
      wp_recv (l x) as "Hl". wp_load. wp_store. by wp_send with "[$Hl]". }
    wp_alloc l as "Hl". wp_send with "[$Hl]". wp_recv as "Hl". wp_load.
    by iApply "HΦ".
  Qed.

End roundtrip_ref.

Definition smuggle_prog : val :=
  λ: <>,
     let: "cs" := new_chan #3 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     Fork (send "c1" #2 (recv "c1" #0));;
     Fork (let: "x" := recv "c2" #1 in send "c2" #0 ("x"+#2));;
     send "c0" #1 #40;; recv "c0" #2.

Section smuggle_ref.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_smuggle : list (iProto Σ) :=
    [(<(Send, 1) @ (x:Z)> MSG #x ; <(Recv,2)> MSG #(x+2) ; END)%proto;
     (<(Recv, 0) @ (v:val)> MSG v ; <(Send,2)> MSG v ; END)%proto;
     (<(Recv, 1) @ (x:Z)> MSG #x ; <(Send,0)> MSG #(x+2) ; END)%proto].

  Lemma iProto_smuggle_consistent :
    ⊢ iProto_consistent iProto_smuggle.
  Proof. rewrite /iProto_smuggle. iProto_consistent_take_steps. Qed.

  Lemma smuggle_spec :
    {{{ True }}} smuggle_prog #() {{{ RET #42 ; True }}}.
  Proof using chanG0 heapGS0 Σ.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_smuggle with iProto_smuggle_consistent
      as (c0 c1 c2) "Hc0" "Hc1" "Hc2".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>". wp_recv (v) as "_". wp_send with "[//]". done. }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>". wp_recv (x) as "_". by wp_send with "[//]". }
    wp_send with "[//]". wp_recv as "_". by iApply "HΦ".
  Qed.

End smuggle_ref.

Definition smuggle_ref_prog : val :=
  λ: <>,
     let: "cs" := new_chan #3 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     Fork (send "c1" #2 (recv "c1" #0));;
     Fork (let: "l" := recv "c2" #1 in "l" <- !"l" + #2;; send "c2" #0 #());;
     let: "l" := ref #40 in
     send "c0" #1 "l";; recv "c0" #2;; !"l".

Section smuggle_ref.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_smuggle_ref : list (iProto Σ) :=
    [(<(Send, 1) @ (l:loc) (x:Z)> MSG #l {{ l ↦ #x }} ;
            <(Recv,2)> MSG #() {{ l ↦ #(x+2) }} ; END)%proto;
     (<(Recv, 0) @ (v:val)> MSG v ; <(Send,2)> MSG v; END)%proto;
     (<(Recv, 1) @ (l:loc) (x:Z)> MSG #l {{ l ↦ #x }} ;
            <(Send,0)> MSG #() {{ l ↦ #(x+2) }} ; END)%proto].

  Lemma iProto_smuggle_ref_consistent :
    ⊢ iProto_consistent iProto_smuggle_ref.
  Proof. rewrite /iProto_smuggle_ref. iProto_consistent_take_steps. Qed.

  Lemma smuggle_ref_spec :
    {{{ True }}} smuggle_ref_prog #() {{{ RET #42 ; True }}}.
  Proof using chanG0 heapGS0 Σ.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_smuggle_ref with iProto_smuggle_ref_consistent
      as (c0 c1 c2) "Hc0" "Hc1" "Hc2".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>". wp_recv (v) as "_". by wp_send with "[//]". }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>". wp_recv (l x) as "Hl". wp_load. wp_store.
      by wp_send with "[$Hl]". }
    wp_alloc l as "Hl". wp_send with "[$Hl]". wp_recv as "Hl". wp_load.
    by iApply "HΦ".
  Qed.

End smuggle_ref.

Definition smuggle_ref_rec_prog : val :=
  λ: <>,
     let: "cs" := new_chan #3 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     Fork ((rec: "go" "c1" :=
              send "c1" #2 (recv "c1" #0);;
                   "go" "c1") "c1");;
     Fork ((rec: "go" "c2" :=
              let: "l" := recv "c2" #1 in "l" <- !"l" + #2;; send "c2" #0 #();;
           "go" "c2") "c2");;
     let: "l" := ref #38 in
     send "c0" #1 "l";; recv "c0" #2;;
     send "c0" #1 "l";; recv "c0" #2;; !"l".

Section smuggle_ref.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_smuggle_ref_rec1_aux (rec : iProto Σ) : iProto Σ :=
    (<(Send, 1) @ (l:loc) (x:Z)> MSG #l {{ l ↦ #x }} ;
     <(Recv,2)> MSG #() {{ l ↦ #(x+2) }} ; rec)%proto.
  Instance iProto_smuggle_ref_rec1_contractive :
    Contractive iProto_smuggle_ref_rec1_aux.
  Proof. solve_proto_contractive. Qed.
  Definition iProto_smuggle_ref_rec1 :=
    fixpoint iProto_smuggle_ref_rec1_aux.
  Lemma iProto_smuggle_ref_rec1_unfold :
    iProto_smuggle_ref_rec1 ≡
                (iProto_smuggle_ref_rec1_aux iProto_smuggle_ref_rec1).
  Proof. apply (fixpoint_unfold _). Qed.
  Global Instance iProto_smuggle_ref_rec1_proto_unfold :
    ProtoUnfold iProto_smuggle_ref_rec1
                (iProto_smuggle_ref_rec1_aux iProto_smuggle_ref_rec1).
  Proof. apply proto_unfold_eq, (fixpoint_unfold _). Qed.

  Definition iProto_smuggle_ref_rec2_aux (rec : iProto Σ) : iProto Σ :=
    (<(Recv, 0) @ (v:val)> MSG v ; <(Send,2)> MSG v; rec)%proto.
  Instance iProto_smuggle_ref_rec2_contractive :
    Contractive iProto_smuggle_ref_rec2_aux.
  Proof. solve_proto_contractive. Qed.
  Definition iProto_smuggle_ref_rec2 :=
    fixpoint iProto_smuggle_ref_rec2_aux.
  Lemma iProto_smuggle_ref_rec2_unfold :
    iProto_smuggle_ref_rec2 ≡
                (iProto_smuggle_ref_rec2_aux iProto_smuggle_ref_rec2).
  Proof. apply (fixpoint_unfold _). Qed.

  Global Instance iProto_smuggle_ref_rec2_proto_unfold :
    ProtoUnfold iProto_smuggle_ref_rec2
                (iProto_smuggle_ref_rec2_aux iProto_smuggle_ref_rec2).
  Proof. apply proto_unfold_eq, (fixpoint_unfold _). Qed.
  Definition iProto_smuggle_ref_rec3_aux (rec : iProto Σ) : iProto Σ :=
    (<(Recv, 1) @ (l:loc) (x:Z)> MSG #l {{ l ↦ #x }} ;
     <(Send,0)> MSG #() {{ l ↦ #(x+2) }} ; rec)%proto.
  Instance iProto_smuggle_ref_rec3_contractive :
    Contractive iProto_smuggle_ref_rec3_aux.
  Proof. solve_proto_contractive. Qed.
  Definition iProto_smuggle_ref_rec3 :=
    fixpoint iProto_smuggle_ref_rec3_aux.
  Lemma iProto_smuggle_ref_rec3_unfold :
    iProto_smuggle_ref_rec3 ≡
                (iProto_smuggle_ref_rec3_aux iProto_smuggle_ref_rec3).
  Proof. apply (fixpoint_unfold _). Qed.
  Global Instance iProto_smuggle_ref_rec3_proto_unfold :
    ProtoUnfold iProto_smuggle_ref_rec3
                (iProto_smuggle_ref_rec3_aux iProto_smuggle_ref_rec3).
  Proof. apply proto_unfold_eq, (fixpoint_unfold _). Qed.

  Definition iProto_smuggle_ref_rec : list (iProto Σ) :=
    [iProto_smuggle_ref_rec1;
     iProto_smuggle_ref_rec2;
     iProto_smuggle_ref_rec3].

  Lemma iProto_smuggle_ref_rec_consistent :
    ⊢ iProto_consistent iProto_smuggle_ref_rec.
  Proof.
    iLöb as "IH".
    rewrite /iProto_smuggle_ref_rec.
    iEval (rewrite iProto_smuggle_ref_rec1_unfold).
    iEval (rewrite iProto_smuggle_ref_rec2_unfold).
    iEval (rewrite iProto_smuggle_ref_rec3_unfold).
    iProto_consistent_take_steps.
    rewrite iProto_smuggle_ref_rec2_unfold.
    iProto_consistent_take_steps.
    rewrite iProto_smuggle_ref_rec2_unfold.
    iProto_consistent_take_steps.
    done.
  Qed.

  Lemma smuggle_ref_rec_prog_spec :
    {{{ True }}} smuggle_ref_rec_prog #() {{{ RET #42 ; True }}}.
  Proof using chanG0.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_smuggle_ref_rec with iProto_smuggle_ref_rec_consistent
      as (c0 c1 c2) "Hc0" "Hc1" "Hc2".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>". wp_pures. iLöb as "IH".
      wp_recv (v) as "_". wp_send with "[//]".
      wp_pures. by iApply "IH". }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>". wp_pure _. iLöb as "IH".
      wp_recv (l x) as "Hl". wp_load. wp_store. wp_send with "[$Hl]".
      do 2 wp_pure _. by iApply "IH". }
    wp_alloc l as "Hl". wp_send with "[$Hl]". wp_recv as "Hl".
    wp_send with "[$Hl]". wp_recv as "Hl". wp_load.
    by iApply "HΦ".
  Qed.

End smuggle_ref.
