(** This file includes exercise examples that each exercise a unique feature of
multiparty dependent separation protocols and Multris. *)
From multi_actris.channel Require Import proofmode.
Set Default Proof Using "Type".

Definition reroute_prog : val :=
  λ: <>,
     let: "cs" := new_chan #4 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     let: "c3" := get_chan "cs" #3 in
     Fork (let: "bv" := recv "c1" #0 in
           send "c1" (if: Fst "bv" then #2 else #3) (Snd "bv"));;
     Fork (let: "x" := recv "c2" #1 in send "c2" #0 ("x" + #2));;
     Fork (let: "x" := recv "c3" #1 in send "c3" #0 ("x" + #2));;
     send "c0" #1 (#true, #40);; recv "c0" #2.

Section reroute.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_reroute : list (iProto Σ) :=
    [(<(Send, 1) @ (b:bool) (x:Z)> MSG (#b,#x);
      <(Recv, if b then 2 else 3) > MSG #(x+2) ; END)%proto;
     END%proto; (* Exercise: replace END with the correct protocol *)
     (<(Recv, 1) @ (x:Z)> MSG #x;
      <(Send, 0)> MSG #(x+2); END)%proto;
     (<(Recv, 1) @ (x:Z)> MSG #x;
      <(Send, 0)> MSG #(x+2); END)%proto].

  Lemma iProto_reroute_consistent :
    ⊢ iProto_consistent iProto_reroute.
  Proof. Admitted.    
    (** Given the correct protocol the following consistency proof should work. *)
    (* rewrite /iProto_reroute. *)
    (* iProto_consistent_take_steps. *)
    (* destruct x; iProto_consistent_take_steps. *)
  (* Qed. *)

  Lemma reroute_prog_spec :
    {{{ True }}} reroute_prog #() {{{ RET #42 ; True }}}.
  Proof using chanG0.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_reroute with iProto_reroute_consistent
      as (c0 c1 c2 c3) "Hc0" "Hc1" "Hc2" "Hc3".
    wp_smart_apply (wp_fork with "[Hc1]").
    { (* Exercise: Prove the rerouting thread using the Coq/Multris tactics. *)
      admit. }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>". wp_recv (x) as "_". by wp_send with "[//]". }
    wp_smart_apply (wp_fork with "[Hc3]").
    { iIntros "!>". wp_recv (x) as "_". by wp_send with "[//]". }
    wp_send with "[//]". wp_recv as "_". by iApply "HΦ".
  Admitted.
  (* Qed. *)

End reroute.

Definition reroute_ref_prog : val :=
  λ: <>,
     let: "cs" := new_chan #4 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     let: "c3" := get_chan "cs" #3 in
     Fork (let: "bv" := recv "c1" #0 in
           send "c1" (if: Fst "bv" then #2 else #3) (Snd "bv"));;
     Fork (let: "l" := recv "c2" #1 in "l" <- !"l" + #2;; send "c2" #0 #());;
     Fork (let: "l" := recv "c3" #1 in "l" <- !"l" + #2;; send "c3" #0 #());;
     let: "l" := ref #40 in
     send "c0" #1 (#true, "l");; recv "c0" #2;; !"l".

Section reroute_ref.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_reroute_ref : list (iProto Σ) :=
    [END%proto; (* Exercise; Replace END with the correct protocol *)
     (<(Recv, 0) @ (b:bool) (v:val)> MSG (#b,v) ;
      <(Send, if b then 2 else 3)> MSG v ; END)%proto;
     (<(Recv, 1) @ (l:loc) (x:Z)> MSG #l {{ l ↦ #x }};
      <(Send, 0)> MSG #() {{ l ↦ #(x + 2) }}; END)%proto;
     (<(Recv, 1) @ (l:loc) (x:Z)> MSG #l {{ l ↦ #x }};
      <(Send, 0)> MSG #() {{ l ↦ #(x + 2) }}; END)%proto].

  Lemma iProto_reroute_ref_consistent :
    ⊢ iProto_consistent iProto_reroute_ref.
  Proof.
    (* Exercise: Give the correct consistency proof. *)
  Admitted.
  
  Lemma reroute_ref_prog_spec :
    {{{ True }}} reroute_ref_prog #() {{{ RET #42 ; True }}}.
  Proof using chanG0.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_reroute_ref with iProto_reroute_ref_consistent
      as (c0 c1 c2 c3) "Hc0" "Hc1" "Hc2" "Hc3".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>". wp_recv (b v) as "_". destruct b; by wp_send with "[//]". }
    wp_smart_apply (wp_fork with "[Hc2]").
    { iIntros "!>". wp_recv (l x) as "Hl". wp_load. wp_store. by wp_send with "[$Hl]". }
    wp_smart_apply (wp_fork with "[Hc3]").
    { iIntros "!>". wp_recv (l x) as "Hl". wp_load. wp_store. by wp_send with "[$Hl]". }
    admit.
    (* Exercise: Prove the main thread using the Coq/Multris tactics. *)
  Admitted.

End reroute_ref.

Definition reroute_ref_rec_prog : val :=
  λ: <>,
     let: "cs" := new_chan #4 in
     let: "c0" := get_chan "cs" #0 in
     let: "c1" := get_chan "cs" #1 in
     let: "c2" := get_chan "cs" #2 in
     let: "c3" := get_chan "cs" #3 in
     Fork ((rec: "go" "c2" :=
              let: "bv" := recv "c1" #0 in
              send "c1" (if: Fst "bv" then #2 else #3) (Snd "bv");;
              "go" "c2") "c2");;
     Fork ((rec: "go" "c2" :=
              let: "l" := recv "c2" #1 in "l" <- !"l" + #2;; send "c2" #0 #();;
           "go" "c2") "c2");;
     Fork ((rec: "go" "c3" :=
              let: "l" := recv "c3" #1 in "l" <- !"l" + #2;; send "c3" #0 #();;
           "go" "c3") "c3");;
     let: "l" := ref #38 in
     send "c0" #1 (#true, "l");; recv "c0" #2;;
     send "c0" #1 (#false, "l");; recv "c0" #3;;
     !"l".

Section reroute_ref_rec.
  Context `{!heapGS Σ, !chanG Σ}.

  Definition iProto_reroute_ref_rec1_aux (rec : iProto Σ) : iProto Σ :=
    (<(Send, 1) @ (b:bool) (l:loc) (x:Z)> MSG (#b,#l) {{ l ↦ #x }};
      <(Recv, if b then 2 else 3) > MSG #() {{ l ↦ #(x + 2) }} ; rec)%proto.
  Instance iProto_reroute_ref_rec1_contractive :
    Contractive iProto_reroute_ref_rec1_aux.
  Proof. solve_proto_contractive. Qed.
  Definition iProto_reroute_ref_rec1 :=
    fixpoint iProto_reroute_ref_rec1_aux.
  Lemma iProto_reroute_ref_rec1_unfold :
    iProto_reroute_ref_rec1 ≡
                (iProto_reroute_ref_rec1_aux iProto_reroute_ref_rec1).
  Proof. apply (fixpoint_unfold _). Qed.
  Global Instance iProto_reroute_ref_rec1_proto_unfold :
    ProtoUnfold iProto_reroute_ref_rec1
                (iProto_reroute_ref_rec1_aux iProto_reroute_ref_rec1).
  Proof. apply proto_unfold_eq, (fixpoint_unfold _). Qed.

  Definition iProto_reroute_ref_rec2_aux (rec : iProto Σ) : iProto Σ :=
    (<(Recv, 0) @ (b:bool) (v:val)> MSG (#b,v) ;
      <(Send, if b then 2 else 3)> MSG v ; rec)%proto.
  Instance iProto_reroute_ref_rec2_contractive :
    Contractive iProto_reroute_ref_rec2_aux.
  Proof. solve_proto_contractive. Qed.
  Definition iProto_reroute_ref_rec2 :=
    fixpoint iProto_reroute_ref_rec2_aux.
  Lemma iProto_reroute_ref_rec2_unfold :
    iProto_reroute_ref_rec2 ≡
                (iProto_reroute_ref_rec2_aux iProto_reroute_ref_rec2).
  Proof. apply (fixpoint_unfold _). Qed.
  Global Instance iProto_reroute_ref_rec2_proto_unfold :
    ProtoUnfold iProto_reroute_ref_rec2
                (iProto_reroute_ref_rec2_aux iProto_reroute_ref_rec2).
  Proof. apply proto_unfold_eq, (fixpoint_unfold _). Qed.

  (* Exercise: Replace END with the right protocol *)
  (* Hint: You need similar boilerplate as the prior recursive protocols *)
  Definition iProto_reroute_ref_rec3 : iProto Σ :=
    END%proto.
  (* Exercise: Prove the lemma and typeclass instance similar to the ones for
               the above recursive protocols. *)

  Definition iProto_reroute_ref_rec : list (iProto Σ) :=
    [iProto_reroute_ref_rec1;
     iProto_reroute_ref_rec2;
     iProto_reroute_ref_rec3;
     iProto_reroute_ref_rec3].

  Lemma iProto_reroute_ref_rec_consistent :
    ⊢ iProto_consistent iProto_reroute_ref_rec.
  Proof. Admitted.
  (** Given the correct protocol and lemmas the following proof script should work *)
  (*   iLöb as "IH". *)
  (*   rewrite /iProto_reroute_ref_rec. *)
  (*   iEval (rewrite iProto_reroute_ref_rec1_unfold). *)
  (*   iEval (rewrite iProto_reroute_ref_rec2_unfold). *)
  (*   iEval (rewrite iProto_reroute_ref_rec3_unfold). *)
  (*   iProto_consistent_take_steps. *)
  (*   destruct x. *)
  (*   - rewrite iProto_reroute_ref_rec2_unfold. *)
  (*     iProto_consistent_take_steps. *)
  (*     rewrite {2}iProto_reroute_ref_rec3_unfold. *)
  (*     done. *)
  (*   - rewrite iProto_reroute_ref_rec2_unfold. *)
  (*     iProto_consistent_take_steps. *)
  (*     rewrite {1}iProto_reroute_ref_rec3_unfold. *)
  (*     done. *)
  (* Qed. *)

  Lemma reroute_ref_rec_prog_spec :
    {{{ True }}} reroute_ref_rec_prog #() {{{ RET #42 ; True }}}.
  Proof using chanG0.
    iIntros (Φ) "_ HΦ". wp_lam.
    wp_new_chan iProto_reroute_ref_rec with iProto_reroute_ref_rec_consistent
      as (c0 c1 c2 c3) "Hc0" "Hc1" "Hc2" "Hc3".
    wp_smart_apply (wp_fork with "[Hc1]").
    { iIntros "!>". wp_pures. iLöb as "IH".
      wp_recv (b v) as "_". destruct b.
      - wp_send with "[//]". wp_pures. by iApply "IH".
      - wp_send with "[//]". wp_pures. by iApply "IH". }
    wp_smart_apply (wp_fork with "[Hc2]").
    { (* Exercise: Prove the recursive sum thread(s) *) 
      admit. }
    wp_smart_apply (wp_fork with "[Hc3]").
    { (* Exercise: Prove the recursive sum thread(s) *) 
      admit. }
    wp_alloc l as "Hl". wp_send with "[$Hl]". wp_recv as "Hl". 
    wp_send with "[$Hl]". wp_recv as "Hl". wp_load. by iApply "HΦ".
  Admitted.
  (* Qed. *)

End reroute_ref_rec.
